#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include <X11/Xatom.h>
#include <X11/extensions/Xrender.h>

#include <compiz.h>


#define NEG_TOGGLE_KEY_DEFAULT              "n"
#define NEG_TOGGLE_MODIFIERS_DEFAULT        CompSuperMask

#define NEG_TOGGLE_ALL_KEY_DEFAULT          "m"
#define NEG_TOGGLE_ALL_MODIFIERS_DEFAULT    CompSuperMask

#define NEG_SCREEN_OPTION_WINDOW_TOGGLE     0
#define NEG_SCREEN_OPTION_SCREEN_TOGGLE     1
#define NEG_SCREEN_OPTION_NUM               2

static int displayPrivateIndex;

typedef struct _NEGDisplay {
    int		    screenPrivateIndex;
    HandleEventProc handleEvent;
} NEGDisplay;


typedef struct _NEGSCreen {
    int                     windowPrivateIndex;
    CompOption              opt[NEG_SCREEN_OPTION_NUM];
    DrawWindowTextureProc   drawWindowTexture;
    DamageWindowRectProc    damageWindowRect;
    Bool                    isNeg;
} NEGScreen;

typedef struct _NEGWindow {
    Bool isNeg;
} NEGWindow;

#define GET_NEG_DISPLAY(d) ((NEGDisplay *) (d)->privates[displayPrivateIndex].ptr)
#define NEG_DISPLAY(d) NEGDisplay *nd = GET_NEG_DISPLAY (d)
#define GET_NEG_SCREEN(s, nd) ((NEGScreen *) (s)->privates[(nd)->screenPrivateIndex].ptr)
#define NEG_SCREEN(s) NEGScreen *ns = GET_NEG_SCREEN (s, GET_NEG_DISPLAY (s->display))
#define GET_NEG_WINDOW(w, ns) ((NEGWindow *) (w)->privates[(ns)->windowPrivateIndex].ptr)
#define NEG_WINDOW(w) NEGWindow *nw = GET_NEG_WINDOW  (w, GET_NEG_SCREEN  (w->screen, GET_NEG_DISPLAY (w->screen->display)))

#define NUM_OPTIONS(s) (sizeof ((s)->opt) / sizeof (CompOption))

static CompOption *
NEGGetScreenOptions (CompScreen *screen, int *count)
{
    NEG_SCREEN (screen);

    *count = NUM_OPTIONS (ns);
    return ns->opt;
}

static Bool
NEGSetScreenOption (CompScreen      *screen,
		       char	       *name,
		       CompOptionValue *value)
{
    CompOption *o;
    int	       index;

    NEG_SCREEN (screen);

    o = compFindOption (ns->opt, NUM_OPTIONS (ns), name, &index);
    if (!o)
	return FALSE;

    switch (index) {
        case NEG_SCREEN_OPTION_WINDOW_TOGGLE:
        case NEG_SCREEN_OPTION_SCREEN_TOGGLE:
            if (addScreenBinding (screen, &value->bind))
	    {
	        removeScreenBinding (screen, &o->value.bind);

	        if (compSetBindingOption (o, value))
		    return TRUE;
	    }
	    break;
 
        default:
            break;
    }
    return FALSE;
}

static void
NEGScreenInitOptions (NEGScreen *ns, Display *display)
{
    CompOption *o;

    o = &ns->opt[NEG_SCREEN_OPTION_WINDOW_TOGGLE];
    o->name                       = "window_toggle";
    o->shortDesc                  = "Toggle Negative";
    o->longDesc                   = "Toggle Window Negative";
    o->type                       = CompOptionTypeBinding;
    o->value.bind.type		  = CompBindingTypeKey;
    o->value.bind.u.key.modifiers = NEG_TOGGLE_MODIFIERS_DEFAULT;
    o->value.bind.u.key.keycode   = XKeysymToKeycode (display, XStringToKeysym (NEG_TOGGLE_KEY_DEFAULT));

    o = &ns->opt[NEG_SCREEN_OPTION_SCREEN_TOGGLE];
    o->name                       = "screen_toggle";
    o->shortDesc                  = "Toggle Negative";
    o->longDesc                   = "Toggle Screen Negative";
    o->type                       = CompOptionTypeBinding;
    o->value.bind.type		  = CompBindingTypeKey;
    o->value.bind.u.key.modifiers = NEG_TOGGLE_ALL_MODIFIERS_DEFAULT;
    o->value.bind.u.key.keycode   = XKeysymToKeycode (display, XStringToKeysym (NEG_TOGGLE_ALL_KEY_DEFAULT));
}

static void
NEGToggle (CompWindow *w)
{
    NEG_WINDOW (w);
    nw->isNeg = !nw->isNeg;
    addWindowDamage (w);
}

static void
NEGToggleScreen (CompScreen *s)
{
    CompWindow *w;
    NEG_SCREEN (s);

    ns->isNeg = !ns->isNeg;
    for (w = s->windows; w; w = w->next)
        if (w) NEGToggle (w);
}

static void
NEGHandleEvent (CompDisplay *d, XEvent *event)
{
    CompScreen *s;
    CompWindow *w;

    NEG_DISPLAY (d);


    switch (event->type) {
        case KeyPress:
            s = findScreenAtDisplay (d, event->xkey.root);
            if (s)
            {
                NEG_SCREEN (s);

                if (eventMatches (d, event, &ns->opt[NEG_SCREEN_OPTION_WINDOW_TOGGLE])) {
                    w = findWindowAtDisplay (d, d->activeWindow);
                    if (w) NEGToggle (w);
                }

                if (eventMatches (d, event, &ns->opt[NEG_SCREEN_OPTION_SCREEN_TOGGLE])) {
                    NEGToggleScreen (s);
                }

            }
            break;

        case ButtonPress:
            s = findScreenAtDisplay (d, event->xkey.root);
            if (s)
            {
                NEG_SCREEN (s);

                if (eventMatches (d, event, &ns->opt[NEG_SCREEN_OPTION_WINDOW_TOGGLE])) {
                    w = findWindowAtDisplay (d, d->activeWindow);
                    if (w) NEGToggle (w);
                }

                if (eventMatches (d, event, &ns->opt[NEG_SCREEN_OPTION_SCREEN_TOGGLE])) {
                    NEGToggleScreen (s);
                }
            }
            break;
    }

    UNWRAP (nd, d, handleEvent);
    (*d->handleEvent) (d, event);
    WRAP (nd, d, handleEvent, NEGHandleEvent);
}

static void
NEGDrawWindowTexture (CompWindow		*w,
  		      CompTexture		*texture,
		      const WindowPaintAttrib   *attrib,
		      unsigned int		mask)
{
    GLfloat constant[4];

    NEG_SCREEN (w->screen);
    NEG_WINDOW (w);

    if (!nw->isNeg || mask & PAINT_WINDOW_DECORATION_MASK)
    {
        UNWRAP (ns, w->screen, drawWindowTexture);
        (*w->screen->drawWindowTexture) (w, texture, attrib, mask);
        WRAP (ns, w->screen, drawWindowTexture, NEGDrawWindowTexture);
    }
    else
    {
        glPushMatrix ();

        if (mask & PAINT_WINDOW_TRANSFORMED_MASK)
        {
	    glTranslatef (w->attrib.x, w->attrib.y, 0.0f);
	    glScalef (attrib->xScale, attrib->yScale, 0.0f);
            glTranslatef (-w->attrib.x, -w->attrib.y, 0.0f);
        }

        enableTexture (w->screen, &w->texture, w->texture.filter);

        glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
        glTexEnvf (GL_TEXTURE_ENV, GL_COMBINE_RGB, GL_REPLACE);
        glTexEnvf (GL_TEXTURE_ENV, GL_SOURCE0_RGB, GL_TEXTURE0_ARB);
        glTexEnvf (GL_TEXTURE_ENV, GL_OPERAND0_RGB, GL_ONE_MINUS_SRC_COLOR);

        if (attrib->opacity < OPAQUE || attrib->brightness != BRIGHT)
        {
            glEnable (GL_BLEND);

            constant[3] = attrib->opacity / 65535.0f;
    	    constant[0] = constant[1] = constant[2] = constant[3] * attrib->brightness / 65535.0f;
            glTexEnvfv (GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, constant);

            glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
	    glTexEnvf (GL_TEXTURE_ENV, GL_COMBINE_RGB, GL_MODULATE);
	    glTexEnvf (GL_TEXTURE_ENV, GL_SOURCE0_RGB, GL_TEXTURE0_ARB);
	    glTexEnvf (GL_TEXTURE_ENV, GL_SOURCE1_RGB, GL_CONSTANT);
	    glTexEnvf (GL_TEXTURE_ENV, GL_OPERAND0_RGB, GL_ONE_MINUS_SRC_COLOR);
	    glTexEnvf (GL_TEXTURE_ENV, GL_OPERAND1_RGB, GL_SRC_COLOR);

	    glTexEnvf (GL_TEXTURE_ENV, GL_COMBINE_ALPHA, GL_MODULATE);
	    glTexEnvf (GL_TEXTURE_ENV, GL_SOURCE0_ALPHA, GL_PREVIOUS);
	    glTexEnvf (GL_TEXTURE_ENV, GL_SOURCE1_ALPHA, GL_CONSTANT);
    	    glTexEnvf (GL_TEXTURE_ENV, GL_OPERAND0_ALPHA, GL_SRC_ALPHA);
	    glTexEnvf (GL_TEXTURE_ENV, GL_OPERAND1_ALPHA, GL_SRC_ALPHA);

	    (*w->screen->drawWindowGeometry) (w);

            glDisable (GL_BLEND);
        }
        else
        {
    	    (*w->screen->drawWindowGeometry) (w);
        }

        disableTexture (w->screen, &w->texture);
        screenTexEnvMode (w->screen, GL_REPLACE);

        glPopMatrix ();
    }
}

static Bool 
NEGDamageWindowRect (CompWindow *w, Bool initial, BoxPtr rect)
{
    int status;

    NEG_SCREEN (w->screen);
    NEG_WINDOW (w);

    if (initial)
    {
        if (ns->isNeg && !nw->isNeg)
            NEGToggle (w);
    }

    UNWRAP (ns, w->screen, damageWindowRect);
    status = (*w->screen->damageWindowRect) (w, initial, rect);
    WRAP (ns, w->screen, damageWindowRect, NEGDamageWindowRect);

    return status;
}

static Bool
NEGInitDisplay (CompPlugin  *p, CompDisplay *d)
{
    NEGDisplay *nd;

    nd = malloc (sizeof (NEGDisplay));
    if (!nd)
	return FALSE;

    nd->screenPrivateIndex = allocateScreenPrivateIndex (d);
    if (nd->screenPrivateIndex < 0)
    {
	free (nd);
	return FALSE;
    }

    WRAP (nd, d, handleEvent, NEGHandleEvent);

    d->privates[displayPrivateIndex].ptr = nd;

    return TRUE;
}

static void
NEGFiniDisplay (CompPlugin  *p, CompDisplay *d)
{
    NEG_DISPLAY (d);

    freeScreenPrivateIndex (d, nd->screenPrivateIndex);

    UNWRAP (nd, d, handleEvent);

    free (nd);
}

static Bool
NEGInitScreen (CompPlugin *p, CompScreen *s)
{
    NEGScreen *ns;
    NEG_DISPLAY (s->display);

    ns = malloc (sizeof (NEGScreen));
    if (!ns)
	return FALSE;

    ns->windowPrivateIndex = allocateWindowPrivateIndex (s);
    if (ns->windowPrivateIndex < 0)
    {
	free (ns);
	return FALSE;
    }

    NEGScreenInitOptions (ns, s->display->display);

    ns->isNeg = FALSE;

    addScreenBinding (s, &ns->opt[NEG_SCREEN_OPTION_WINDOW_TOGGLE].value.bind);
    addScreenBinding (s, &ns->opt[NEG_SCREEN_OPTION_SCREEN_TOGGLE].value.bind);
  
    WRAP (ns, s, drawWindowTexture, NEGDrawWindowTexture);
    WRAP (ns, s, damageWindowRect, NEGDamageWindowRect);

    s->privates[nd->screenPrivateIndex].ptr = ns;

    return TRUE;
}


static void
NEGFiniScreen (CompPlugin *p, CompScreen *s)
{
    NEG_SCREEN (s);
    freeWindowPrivateIndex (s, ns->windowPrivateIndex);
    UNWRAP (ns, s, drawWindowTexture);
    UNWRAP (ns, s, damageWindowRect);
    free (ns);
}

static Bool
NEGInitWindow (CompPlugin *p, CompWindow *w)
{
    NEGWindow *nw;

    NEG_SCREEN (w->screen);

    nw = malloc (sizeof (NEGWindow));
    if (!nw)
	return FALSE;

    nw->isNeg = FALSE;

    w->privates[ns->windowPrivateIndex].ptr = nw;

    return TRUE;
}

static void
NEGFiniWindow (CompPlugin *p, CompWindow *w)
{
    NEG_WINDOW (w);
    free (nw);
}

static Bool
NEGInit (CompPlugin *p)
{
    displayPrivateIndex = allocateDisplayPrivateIndex ();
    if (displayPrivateIndex < 0)
	return FALSE;

    return TRUE;
}

static void
NEGFini (CompPlugin *p)
{
    if (displayPrivateIndex >= 0)
	freeDisplayPrivateIndex (displayPrivateIndex);
}

CompPluginVTable NEGVTable = {
    "neg",
    "Negative",
    "Negative",
    NEGInit,
    NEGFini,
    NEGInitDisplay,
    NEGFiniDisplay,
    NEGInitScreen,
    NEGFiniScreen,
    NEGInitWindow,
    NEGFiniWindow,
    0, /* GetDisplayOptions */
    0, /* SetDisplayOption */
    NEGGetScreenOptions,
    NEGSetScreenOption,
    0,
    0
};

CompPluginVTable *
getCompPluginInfo (void)
{
    return &NEGVTable;
}



