
#ifdef HAVE_CONFIG_H
#  include "../config.h"
#endif

#include "quinn.h"

#include <stdlib.h>
#include <string.h>

#define DBUS_API_SUBJECT_TO_CHANGE

#include <compiz.h>
#include <dbus/dbus.h>

CompDbusContext *compDbusContext = NULL;

extern Bool disableDbus;

static DBusHandlerResult
dbusFilterMessage ( DBusConnection *connection,
		    DBusMessage *message,
		    void *data );

static void
coreDbusFilter ( CompScreen *screen,
		 CompDbusMessage *request );

static void
dbusIntrospect ( DBusMessage *request );

Bool
dbusInit (  )
{
	CompDbusContext *context = compDbusContext;
	if ( context == NULL )
	{
		compDbusContext = context = malloc ( sizeof ( CompDbusContext ) );
		if ( context == NULL)
		{
			fprintf ( stderr, "%s: Couldn't allocate enough memory\n",
		 		  programName );
			return FALSE;
		}

		context->screens = context->reverseScreens = NULL;
		context->registered = FALSE;
		context->connection = NULL;
	
		if ( !dbusGetConnection ( context ) )
		{
			fprintf ( stderr, "%s: Disabling D-BUS Service\n",
				  programName );
			dbusFini ( );
			return FALSE;
		}
	}
	return TRUE;
}

Bool
dbusFini (  )
{
	CompDbusContext *context = compDbusContext;
	if ( context != NULL )
	{
		CompDbusScreen *ds, *dsn;
		ds = context->screens;
		while ( ds )
		{
			dsn = ds->next;
			dbusFiniScreen ( ds->screen );
			ds = dsn;
		}
		context->screens = NULL;
		if ( context->registered )
		{
			dbus_connection_remove_filter ( context->connection, dbusFilterMessage, NULL );
			context->registered = FALSE;
		}
		if ( context->connection != NULL )
		{
			dbus_bus_release_name ( context->connection, COMPIZ_DBUS_SERVICE_NAME, NULL );
			dbus_connection_close ( context->connection );
		}
		free ( context );
		compDbusContext = context = NULL;
		disableDbus = TRUE;
	}
	return TRUE;
}

Bool
dbusGetConnection ( CompDbusContext *context )
{
	DBusError error;
	int ret;
	
	dbus_error_init( &error );

	context->connection = dbus_bus_get ( DBUS_BUS_SESSION, &error );
	if ( dbus_error_is_set( &error ) )
	{
		fprintf ( stderr, "%s: Connection Error (%s)\n", programName, error.message ); 
		dbus_error_free( &error ); 
	}
	if ( context->connection == NULL )
	{
		fprintf ( stderr, "%s: Couldn't initiate D-BUS connection\n",
	 		  programName );
		return FALSE;
	}
	if ( dbus_error_is_set( &error ) ) dbus_error_free ( &error );
	
	ret = dbus_bus_request_name ( context->connection, COMPIZ_DBUS_SERVICE_NAME,
		DBUS_NAME_FLAG_REPLACE_EXISTING | DBUS_NAME_FLAG_ALLOW_REPLACEMENT, &error );
	if ( dbus_error_is_set ( &error ) )
	{ 
		fprintf( stderr, "%s: Name Error (%s)\n", programName, error.message ); 
		dbus_error_free( &error );
	}
	if ( ret != DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER )
	{ 
		fprintf ( stderr, "%s: Can't own requested name (%d), disabling D-BUS\n",
			  programName, ret );
		return FALSE;
	}

	return TRUE;
}

Bool
dbusCheckConnection (  )
{
	CompDbusContext *context = compDbusContext;
	if ( context == NULL ) return FALSE;
	if ( dbus_connection_get_is_connected ( context->connection ) )
		return TRUE;

	fprintf ( stderr, "%s: D-BUS connection lost, reconnecting\n",
 		  programName );

	context->registered = FALSE;
	dbus_connection_close ( context->connection );
	dbus_connection_unref ( context->connection );
	if ( !dbusGetConnection ( context ) )
	{
		fprintf ( stderr, "%s: Can't reconnect, disabling D-BUS\n",
 			  programName );
		dbusFini (  );
		return FALSE;
	}
	else	
		fprintf ( stderr, "%s: D-BUS reconnected\n",
 			  programName );
 			  
 	return TRUE;
}

CompDbusScreen*
dbusFindScreen ( CompScreen *screen )
{
	CompDbusScreen *ds;
	if ( compDbusContext == NULL ) return NULL;
	for ( ds = compDbusContext->screens; ds; ds = ds->next )
	{
		if ( ds->screen == screen )
			return ds;
	}
	return NULL;
}

CompDbusPlugin*
dbusFindPlugin ( CompScreen *screen,
		 CompPlugin *plugin )
{
	CompDbusScreen *ds;
	CompDbusPlugin *dp;
	ds = dbusFindScreen ( screen );
	if ( ds == NULL ) return NULL;
	for ( dp = ds->plugins; dp; dp = dp->next )
	{
		if ( dp->plugin == plugin )
			return dp;
	}
	return NULL;
}

CompDbusMethod*
dbusFindMethod ( CompScreen *screen,
		 CompPlugin *plugin,
		 const char *method )
{
	CompDbusPlugin *dp = dbusFindPlugin ( screen, plugin );
	if ( dp == NULL ) return NULL;
	CompDbusMethod *dm;
	for ( dm = dp->methods; dm; dm = dm->next )
	{
		if ( strcmp ( dm->methodId, method ) == 0 )
			return dm;
	}
	return NULL;
}

Bool
dbusInitScreen ( CompScreen *screen )
{
	CompDbusContext *context = compDbusContext;
	if ( context == NULL ) return FALSE;
	if ( dbusFindScreen ( screen ) != NULL ) return TRUE;
	CompDbusScreen *ds;
	ds = malloc ( sizeof ( CompDbusScreen ) );
	if ( ds == NULL )
	{
		fprintf ( stderr, "%s: Couldn't allocate enough memory\n",
		 	  programName );
		return FALSE;
	}
	char buffer[50];
	int len;
	len = snprintf ( buffer, 50, "%d", screen->screenNum );
	ds->screenId = malloc ( ( strlen( COMPIZ_DBUS_PATH_BASE ) + len + 1 ) * sizeof ( char ) );
	strcpy ( ds->screenId, COMPIZ_DBUS_PATH_BASE );
	strcat ( ds->screenId, buffer );
	ds->context = context;
	ds->screen = screen;
	ds->prev = context->reverseScreens;
	ds->next = NULL;
	ds->plugins = ds->reversePlugins = NULL;
	ds->pluginsCount = 0;
	if ( context->reverseScreens != NULL ) context->reverseScreens->next = ds;
	else context->screens = ds;
	context->reverseScreens = ds;
	context->screensCount++;
	return dbusInitCorePlugin ( ds );
}

Bool
dbusFiniScreen ( CompScreen *screen )
{
	CompDbusScreen *ds = dbusFindScreen ( screen );
	if ( ds == NULL ) return FALSE;
	CompDbusPlugin *dp, *dpnext;
	for ( dp = ds->plugins; dp; dp = dpnext )
	{
		dpnext = dp->next;
		dbusFiniDbusPlugin ( dp );
	}
	if ( ds->pluginsCount != 0 ) return FALSE;
	return dbusFiniDbusScreen ( ds );
}

Bool
dbusFiniDbusScreen ( CompDbusScreen *ds )
{
	if ( ds == NULL ) return FALSE;
	if ( ds->next != NULL ) ds->next->prev = ds->prev;
	else ds->context->reverseScreens = ds->prev;
	if ( ds->prev != NULL ) ds->prev->next = ds->next;
	else ds->context->screens = ds->next;
	ds->context->screensCount--;
	free ( ds->screenId );
	ds->screenId = NULL;
	free ( ds );
	return TRUE;
}

Bool
dbusInitPlugin ( CompScreen *screen,
		 CompPlugin *plugin )
{
	CompDbusContext *context = compDbusContext;
	if ( context == NULL ) return FALSE;
	CompDbusScreen *ds;
	ds = dbusFindScreen ( screen );
	if ( ds == NULL ) return FALSE;
	if ( dbusFindPlugin ( screen, plugin ) != NULL ) return TRUE;
	CompDbusPlugin *dp;
	dp = malloc ( sizeof ( CompDbusPlugin ) );
	if ( dp == NULL )
	{
		fprintf ( stderr, "%s: Couldn't allocate enough memory\n",
		 	  programName );
		return FALSE;
	}
	if ( plugin == NULL )
	{
		dp->pluginId = malloc ( ( strlen( COMPIZ_DBUS_INTERFACE_BASE ) + 5 ) * sizeof ( char ) );
		strcpy ( dp->pluginId, COMPIZ_DBUS_INTERFACE_BASE );
		strcat ( dp->pluginId, "core" );
	}
	else
	{
		dp->pluginId = malloc ( ( strlen( COMPIZ_DBUS_INTERFACE_BASE ) + strlen( plugin->vTable->name ) + 1 ) * sizeof ( char ) );
		strcpy ( dp->pluginId, COMPIZ_DBUS_INTERFACE_BASE );
		strcat ( dp->pluginId, plugin->vTable->name );
	}
	dp->screen = ds;
	dp->plugin = plugin;
	dp->prev = ds->reversePlugins;
	dp->next = NULL;
	dp->methods = dp->reverseMethods = NULL;
	dp->methodsCount = 0;
	if ( ds->reversePlugins != NULL ) ds->reversePlugins->next = dp;
	else ds->plugins = dp;
	ds->reversePlugins = dp;
	ds->pluginsCount++;
	return TRUE;
}

Bool
dbusFiniPlugin ( CompScreen *screen, CompPlugin *plugin )
{
	CompDbusPlugin *dp = dbusFindPlugin ( screen, plugin );
	if ( dp == NULL ) return FALSE;
	CompDbusMethod *dm, *dmnext;
	for ( dm = dp->methods; dm; dm = dmnext )
	{
		dmnext = dm->next;
		dbusUnregisterDbusMethod ( dm );
	}
	if ( dp->methodsCount != 0 ) return FALSE;
	return dbusFiniDbusPlugin ( dp );
}

Bool
dbusFiniDbusPlugin ( CompDbusPlugin *dp )
{
	if ( dp == NULL ) return FALSE;
	if ( dp->next != NULL ) dp->next->prev = dp->prev;
	else dp->screen->reversePlugins = dp->prev;
	if ( dp->prev != NULL ) dp->prev->next = dp->next;
	else dp->screen->plugins = dp->next;
	dp->screen->pluginsCount--;
	free ( dp->pluginId );
	dp->pluginId = NULL;
	free ( dp );	
	return TRUE;
}


Bool
dbusInitCorePlugin ( CompDbusScreen *ds )
{
	if ( !dbusInitPlugin ( ds->screen, NULL ) ) return FALSE;
	dbusRegisterCoreMethod ( ds->screen, "getVersion", 0, coreDbusFilter, TRUE );
	dbusRegisterCoreMethod ( ds->screen, "activateWindow", 1, coreDbusFilter, FALSE );
	return TRUE;
}

Bool dbusRegisterMethod ( CompScreen *screen,
			  CompPlugin *plugin,
			  const char *method,
			  unsigned int argumentsCount,
			  CompDbusHandler *handler,
			  Bool methodSendsReply )
{
	CompDbusPlugin *dp = dbusFindPlugin ( screen, plugin );
	if ( dp == NULL ) return FALSE;
	CompDbusMethod *dm;
	dm = malloc ( sizeof ( CompDbusMethod ) );
	if ( dm == NULL )
	{
		fprintf ( stderr, "%s: Couldn't allocate enough memory\n",
		 	  programName );
		return FALSE;
	}
	dm->methodId = malloc ( ( strlen ( method ) + 1 ) * sizeof ( char ) );
	strcpy ( dm->methodId, method );
	dm->argumentsCount = argumentsCount;
	dm->handler = handler;
	dm->methodSendsReply = methodSendsReply;
	dm->plugin = dp;
	dm->prev = dp->reverseMethods;
	dm->next = NULL;
	if ( dp->reverseMethods != NULL ) dp->reverseMethods->next = dm;
	else dp->methods = dm;
	dp->reverseMethods = dm;
	dp->methodsCount++;
	CompDbusContext *context = dp->screen->context;
	if ( !context->registered )
	{
		context->registered = dbus_connection_add_filter ( context->connection, dbusFilterMessage, context, NULL );	
		if ( !context->registered )
		{
			fprintf ( stderr, "%s: Can't enable message filtering, disabling D-BUS\n",
 				  programName );
			return FALSE;
		}
	}

	return TRUE;
}

Bool
dbusUnregisterMethod (  CompScreen *screen,
		 	CompPlugin *plugin,
			const char *method )
{
	CompDbusMethod *dm = dbusFindMethod ( screen, plugin, method );
	if ( dm == NULL ) return FALSE;
	CompDbusContext *context = dm->plugin->screen->context;
	if ( !dbusUnregisterDbusMethod ( dm ) )
		return FALSE;
	if ( context->screensCount == 0 && context->registered )
	{
		dbus_connection_remove_filter ( context->connection, dbusFilterMessage, NULL );
		context->registered = FALSE;
	}
	return TRUE;
}

Bool
dbusUnregisterDbusMethod ( CompDbusMethod *dm )
{
	if ( dm == NULL ) return FALSE;
	if ( dm->next != NULL ) dm->next->prev = dm->prev;
	else dm->plugin->reverseMethods = dm->prev;
	if ( dm->prev != NULL ) dm->prev->next = dm->next;
	else dm->plugin->methods = dm->next;
	dm->plugin->methodsCount--;
	free ( dm->methodId );
	dm->methodId = NULL;
	free ( dm );
	return TRUE;
}

Bool
dbusRegisterCoreMethod ( CompScreen *screen,
			 const char *method,
			 unsigned int argumentsCount,
			 CompDbusHandler *handler,
			 Bool methodSendsReply )
{
	return dbusRegisterMethod ( screen, NULL, method, argumentsCount, handler, methodSendsReply );
}

Bool
dbusUnregisterCoreMethod ( CompScreen *screen,
			   const char *method )
{
	return dbusUnregisterMethod ( screen, NULL, method );
}

static DBusHandlerResult
dbusFilterMessage ( DBusConnection *connection,
		    DBusMessage *message,
		    void *data )
{
	CompDbusContext *context = data;
	const char *service, *path, *interface, *member;

	service = dbus_message_get_destination ( message );
	path = dbus_message_get_path ( message );
	interface = dbus_message_get_interface ( message );
	member = dbus_message_get_member ( message );
	
	if ( service == NULL || path == NULL ||
	     interface == NULL || member == NULL ||
	     !dbus_message_is_method_call ( message, interface, member ) )
		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
	
	if ( dbus_message_has_interface( message, DBUS_INTERFACE_INTROSPECTABLE ) &&
	     dbus_message_has_member ( message, "Introspect" ) )
	{
		dbusIntrospect ( message );
		return DBUS_HANDLER_RESULT_HANDLED;
	}

	if ( !dbus_message_has_destination ( message, COMPIZ_DBUS_SERVICE_NAME ) )
		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;

	CompDbusScreen *ds;
	for ( ds = context->screens; ds; ds = ds->next )
	{
		if ( strcmp ( ds->screenId, path ) == 0 )
			break;
	}
	if ( !ds )
	{
		dbusErrorMessage ( context->connection, message, COMPIZ_DBUS_ERROR_UNKNOWN_SCREEN, path );
		return DBUS_HANDLER_RESULT_HANDLED;
	}
	
	CompDbusPlugin *dp;
	for ( dp = ds->plugins; dp; dp = dp->next )
	{
		if ( strcmp ( dp->pluginId, interface ) == 0 )
			break;
	}
	if ( !dp )
	{
		dbusErrorMessage ( context->connection, message, COMPIZ_DBUS_ERROR_UNKNOWN_PLUGIN, interface );
		return DBUS_HANDLER_RESULT_HANDLED;
	}
	
	CompDbusMethod *dm;
	for ( dm = dp->methods; dm; dm = dm->next )
	{
		if ( strcmp ( dm->methodId, member ) == 0 )
			break;
	}
	if ( !dm )
	{
		dbusErrorMessage ( context->connection, message, COMPIZ_DBUS_ERROR_UNKNOWN_METHOD, member );
		return DBUS_HANDLER_RESULT_HANDLED;
	}
	
	if ( dm->handler == NULL )
	{
		dbusErrorMessage ( context->connection, message, COMPIZ_DBUS_ERROR_METHOD_NOT_YET_IMPLEMENTED, dm->methodId );
		return DBUS_HANDLER_RESULT_HANDLED;
	}
	
	CompDbusMessage *dbusMessage = dbusBuildCompMessage ( message );

	if ( dbusMessage != NULL )
		dm->handler ( ds->screen, dbusMessage );
	
	return DBUS_HANDLER_RESULT_HANDLED;
}

Bool
dbusErrorMessage ( DBusConnection *connection,
		   DBusMessage *message,
		   const char *error,
		   const char *data )
{
	DBusMessage *reply;
	reply = dbus_message_new_error ( message, error, data );
	dbus_connection_send ( connection, reply, NULL );
	dbus_message_unref ( reply );
	return TRUE;
}

static void
dbusIntrospect ( DBusMessage *request )
{
	CompDbusContext *context = compDbusContext;
	const char *empty = "", *path;
	char *data;
	int length = 0, nodeLength,
			interfaceStartLength, interfaceEndLength,
			methodStartLength, methodEndLength,
			methodReplyLength;
	char buffer[500], buffer2[500];
	int bufferLength;
	
	if ( context == NULL )
		return;

	nodeLength = strlen ( COMPIZ_DBUS_INTROSPECTION_NODE ) - 2;
	interfaceStartLength = strlen ( COMPIZ_DBUS_INTROSPECTION_INTERFACE_START ) - 2;
	interfaceEndLength = strlen ( COMPIZ_DBUS_INTROSPECTION_INTERFACE_END );
	methodStartLength = strlen ( COMPIZ_DBUS_INTROSPECTION_METHOD_START ) - 2;
	methodEndLength = strlen ( COMPIZ_DBUS_INTROSPECTION_METHOD_END );
	methodReplyLength = strlen ( COMPIZ_DBUS_INTROSPECTION_METHOD_REPLY );
	
	length += strlen ( COMPIZ_DBUS_INTROSPECTION_HEADER );

	path = dbus_message_get_path ( request );

	if ( strcmp ( path, "/" ) == 0 )
	{
		length += nodeLength + 3 + strlen ( COMPIZ_DBUS_INTROSPECTION_FOOTER ) + 1;
		data = malloc ( length * sizeof ( char ) );
		if ( data == NULL )
			return;
		strcpy ( data, COMPIZ_DBUS_INTROSPECTION_HEADER );
		bufferLength = snprintf ( buffer, 500, COMPIZ_DBUS_INTROSPECTION_NODE, "org" );
		strcat ( data, buffer );
		strcat ( data, COMPIZ_DBUS_INTROSPECTION_FOOTER );
	}
	else if ( strcmp ( path, "/org" ) == 0 )
	{
		length += nodeLength + 6 + strlen ( COMPIZ_DBUS_INTROSPECTION_FOOTER ) + 1;
		data = malloc ( length * sizeof ( char ) );
		if ( data == NULL )
			return;
		strcpy ( data, COMPIZ_DBUS_INTROSPECTION_HEADER );
		bufferLength = snprintf ( buffer, 500, COMPIZ_DBUS_INTROSPECTION_NODE, "compiz" );
		strcat ( data, buffer );
		strcat ( data, COMPIZ_DBUS_INTROSPECTION_FOOTER );
	}
	else if ( strcmp ( path, "/org/compiz" ) == 0 )
	{
		length += nodeLength + 4 + strlen ( COMPIZ_DBUS_INTROSPECTION_FOOTER ) + 1;
		data = malloc ( length * sizeof ( char ) );
		if ( data == NULL )
			return;
		strcpy ( data, COMPIZ_DBUS_INTROSPECTION_HEADER );
		bufferLength = snprintf ( buffer, 500, COMPIZ_DBUS_INTROSPECTION_NODE, "dbus" );
		strcat ( data, buffer );
		strcat ( data, COMPIZ_DBUS_INTROSPECTION_FOOTER );
	}
	else if ( strcmp ( path, "/org/compiz/dbus" ) == 0 )
	{
		CompDbusScreen *ds;
		for ( ds = context->screens; ds; ds = ds->next )
		{
			bufferLength = snprintf ( buffer, 500, "screen%d", ds->screen->screenNum );
			length += bufferLength + nodeLength;
		}
		length += strlen ( COMPIZ_DBUS_INTROSPECTION_FOOTER ) + 1;
		data = malloc ( length * sizeof ( char ) );
		if ( data == NULL )
			return;
		strcpy ( data, COMPIZ_DBUS_INTROSPECTION_HEADER );
		for ( ds = context->screens; ds; ds = ds->next )
		{
			snprintf ( buffer, 500, "screen%d", ds->screen->screenNum );
			snprintf ( buffer2, 500, COMPIZ_DBUS_INTROSPECTION_NODE, buffer );
			strcat ( data, buffer2 );
		}
		strcat ( data, COMPIZ_DBUS_INTROSPECTION_FOOTER );
	}
	else
	{
		CompDbusScreen *ds;
		CompDbusPlugin *dp;
		CompDbusMethod *dm;
		for ( ds = context->screens; ds; ds = ds->next )
		{
			if ( strcmp ( path, ds->screenId ) != 0 ) continue;
			for ( dp = ds->plugins; dp; dp = dp->next )
			{
				length += strlen ( dp->pluginId ) + interfaceStartLength + interfaceEndLength;
				for ( dm = dp->methods; dm; dm = dm->next )
				{
					length += strlen ( dm->methodId ) + methodStartLength + methodEndLength;
					if ( dm->methodSendsReply == COMPIZ_DBUS_METHOD_SENDS_REPLY )
					{
						length += methodReplyLength;
					}
				}
			}
		}
	
		length += strlen ( COMPIZ_DBUS_INTROSPECTION_FOOTER ) + 1;
		
		data = malloc ( length * sizeof ( char ) );

		if ( data == NULL )
			return;

		strcpy ( data, COMPIZ_DBUS_INTROSPECTION_HEADER );

		for ( ds = context->screens; ds; ds = ds->next )
		{
			if ( strcmp ( path, ds->screenId ) != 0 ) continue;
			for ( dp = ds->plugins; dp; dp = dp->next )
			{
				bufferLength = snprintf ( buffer, 500, COMPIZ_DBUS_INTROSPECTION_INTERFACE_START, dp->pluginId );
				strcat ( data, buffer );
				for ( dm = dp->methods; dm; dm = dm->next )
				{
					bufferLength = snprintf ( buffer, 500, COMPIZ_DBUS_INTROSPECTION_METHOD_START, dm->methodId );
					strcat ( data, buffer );
					if ( dm->methodSendsReply == COMPIZ_DBUS_METHOD_SENDS_REPLY )
					{
						strcat ( data, COMPIZ_DBUS_INTROSPECTION_METHOD_REPLY );
					}
					strcat ( data, COMPIZ_DBUS_INTROSPECTION_METHOD_END );
				}
				strcat ( data, COMPIZ_DBUS_INTROSPECTION_INTERFACE_END );
			}
		}	

		strcat ( data, COMPIZ_DBUS_INTROSPECTION_FOOTER );
	}

	DBusMessage *reply;
	reply = dbus_message_new_method_return ( request );
	if ( data )
		dbus_message_append_args ( reply, DBUS_TYPE_STRING, &data, DBUS_TYPE_INVALID );
	else
		dbus_message_append_args ( reply, DBUS_TYPE_STRING, &empty, DBUS_TYPE_INVALID );
	dbus_connection_send ( context->connection, reply, NULL );
	dbus_message_unref ( reply );
	free( data );
}

Bool
dbusSendMethodCall ( const char *service,
		     const char *path,
		     const char *interface,
		     const char *method,
		     CompDbusArgument *arguments )
{
	CompDbusContext *context = compDbusContext;
	if ( context == NULL ) return FALSE;
	DBusMessage *message;
	DBusMessageIter args;
	dbus_uint32_t serial = 0;
	int integer;
	char *string;

	message = dbus_message_new_method_call ( service, path, interface, method );
	if ( message == NULL )
	{
		fprintf ( stderr, "%s: Couldn't create D-BUS method call", programName );
		return FALSE;
	}
	
	dbus_message_iter_init_append ( message, &args );
	CompDbusArgument *da;
	for ( da = arguments; da; da = da->next )
	{
		if ( da->argumentType & CompDbusArgumentIntMask )
		{
			integer = da->integerValue;
			if ( !dbus_message_iter_append_basic ( &args, DBUS_TYPE_INT32, &integer ) )
			{
				fprintf ( stderr, "%s: Couldn't allocate enough memory for D-BUS method call arguments\n", programName ); 
				return FALSE;
			}
		}
		else if ( da->argumentType & CompDbusArgumentStringMask )
		{
			string = da->stringValue;
			if ( !dbus_message_iter_append_basic ( &args, DBUS_TYPE_STRING, &string ) )
			{
				fprintf ( stderr, "%s: Couldn't allocate enough memory for D-BUS method call arguments\n", programName ); 
				return FALSE;
			}
		}
	}

	if ( !dbus_connection_send ( context->connection, message, &serial ) )
	{
		fprintf ( stderr, "%s: Couldn't send D-BUS method call", programName );
		return FALSE;
	}
	
	dbus_message_unref ( message );

	return TRUE;
}

CompDbusMessage*
dbusBuildCompMessage ( DBusMessage *message )
{
	CompDbusMessage *dbusMessage;
	if ( message == NULL ) return NULL;
	dbusMessage = malloc ( sizeof ( CompDbusMessage ) );
	if ( dbusMessage == NULL ) return NULL;
	dbusMessage->message = message;
	dbus_message_ref ( dbusMessage->message );

	DBusMessageIter iterator;
	int32_t integer;
	char *string;
	Bool more = TRUE;
	
	dbusMessage->arguments = dbusMessage->reverseArguments = NULL;
	CompDbusArgument *da = NULL;

	if ( dbus_message_iter_init( message, &iterator ) )
	{
		while ( more )
		{
			switch ( dbus_message_iter_get_arg_type( &iterator ) )
			{
				case DBUS_TYPE_STRING:
					dbus_message_iter_get_basic( &iterator, &string );
					da = malloc ( sizeof ( CompDbusArgument ) );
					if ( da == NULL ) break;
					da->stringValue = malloc( ( strlen ( string ) + 1 ) * sizeof ( char ) );
					if ( da->stringValue == NULL )
					{
						free ( da );
						break;
					}
					strcpy ( da->stringValue, string );
					da->integerValue = 0;
					da->argumentType = 0;
					da->argumentType |= CompDbusArgumentStringMask;
					if ( dbusMessage->reverseArguments != NULL )
						dbusMessage->reverseArguments->next = da;
					else dbusMessage->arguments = da;
					da->prev = dbusMessage->reverseArguments;
					dbusMessage->reverseArguments = da;
					dbusMessage->argumentsCount++;
					break;
				case DBUS_TYPE_INT32:
					dbus_message_iter_get_basic( &iterator, &integer );
					da = malloc ( sizeof ( CompDbusArgument ) );
					if ( da == NULL ) break;
					da->stringValue = NULL;
					da->integerValue = integer;
					da->argumentType = 0;
					da->argumentType |= CompDbusArgumentIntMask;
					if ( dbusMessage->reverseArguments != NULL )
						dbusMessage->reverseArguments->next = da;
					else dbusMessage->arguments = da;
					da->prev = dbusMessage->reverseArguments;
					dbusMessage->reverseArguments = da;
					dbusMessage->argumentsCount++;
					break;
				case DBUS_TYPE_INVALID:
					more = FALSE;
					break;
				default:
					break;
			}
			if ( !dbus_message_iter_has_next( &iterator ) ||
				!dbus_message_iter_next( &iterator) )
					more = FALSE;
		}
	}

	return dbusMessage;
}

Bool
dbusFreeCompMessage ( CompDbusMessage *dbusMessage )
{
	if ( dbusMessage == NULL ) return TRUE;
	dbus_message_unref ( dbusMessage->message );
	CompDbusArgument *da, *dan;
	for ( da = dbusMessage->arguments; da; da = dan )
	{
		dan = da->next;
		if ( da->argumentType & CompDbusArgumentStringMask )
			free ( da->stringValue );
		free ( da );
	}
	free ( dbusMessage );
	return TRUE;
}

CompDbusArgument*
dbusGetArgument ( CompDbusMessage *dbusMessage,
		  int n )
{
	if ( dbusMessage == NULL ) return NULL;
	if ( n >= dbusMessage->argumentsCount || n < 0 ) return NULL;
	CompDbusArgument *da = dbusMessage->arguments;
	int i;
	for ( i = 0; i < n; i++ )
		da = da->next;
	return da;
}

int
dbusGetArgumentType ( CompDbusMessage *dbusMessage,
		      int n )
{
	CompDbusArgument *da = dbusGetArgument ( dbusMessage, n );
	if ( da == NULL ) return -1;
	if ( da->argumentType & CompDbusArgumentIntMask )
		return COMP_INT_ARG;
	if ( da->argumentType & CompDbusArgumentStringMask )
		return COMP_STRING_ARG;
	return -1;
}

static void
coreDbusFilter ( CompScreen *screen,
		 CompDbusMessage *request )
{
	const char *method = dbus_message_get_member ( request->message );
	if ( strcmp ( method, "getVersion" ) == 0 )
	{
		const char *version = PACKAGE_STRING QUINN_STRING;
		DBusMessage *reply;
		reply = dbus_message_new_method_return ( request->message );
		dbus_message_append_args ( reply, DBUS_TYPE_STRING, &version, DBUS_TYPE_INVALID );
		dbus_connection_send ( compDbusContext->connection, reply, NULL );
		dbus_message_unref ( reply );
	}
	else if (strcmp (method, "activateWindow") == 0)
	{
		if (dbusGetArgumentType (request, 0) == COMP_STRING_ARG) {
			CompDbusArgument *winId = dbusGetArgument(request, 0);
			int id;
			if (winId && sscanf(winId->stringValue,"%x",&id) == 1) {
				CompWindow *targetWindow = findWindowAtScreen (screen, (Window)id);
				if (targetWindow)
					sendWindowActivationRequest (screen, targetWindow->id);
			}
		}
	}
}
